import json
from pathlib import Path

try:  # Python 3.8
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata  # noqa: F401


pkg_meta = importlib_metadata.metadata(__name__.split(".")[0])

configs = {
    'en_bert_base_nli_mean_tokens': {
        'spacy_base_model': 'en',
        'dimensions': 768,
        'name': 'bert-base-nli-mean-tokens'
    },
    'en_bert_base_nli_max_tokens': {
        'spacy_base_model': 'en',
        'dimensions': 768,
        'name': 'bert-base-nli-max-tokens'
    },
    'en_bert_base_nli_cls_token': {
        'spacy_base_model': 'en',
        'dimensions': 768,
        'name': 'bert-base-nli-cls-token'
    },
    'en_bert_large_nli_mean_tokens': {
        'spacy_base_model': 'en',
        'dimensions': 1024,
        'name': 'bert-large-nli-mean-tokens'
    },
    'en_bert_large_nli_max_tokens': {
        'spacy_base_model': 'en',
        'dimensions': 1024,
        'name': 'bert-large-nli-max-tokens'
    },
    'en_bert_large_nli_cls_token': {
        'spacy_base_model': 'en',
        'dimensions': 1024,
        'name': 'bert-large-nli-cls-token'
    },
    'en_roberta_base_nli_mean_tokens': {
        'spacy_base_model': 'en',
        'dimensions': 768,
        'name': 'roberta-base-nli-mean-tokens'
    },
    'en_roberta_large_nli_mean_tokens': {
        'spacy_base_model': 'en',
        'dimensions': 1024,
        'name': 'roberta-large-nli-mean-tokens'
    },
    'en_distilbert_base_nli_mean_tokens': {
        'spacy_base_model': 'en',
        'dimensions': 768,
        'name': 'distilbert-base-nli-mean-tokens'
    },
    'en_bert_base_nli_stsb_mean_tokens': {
        'spacy_base_model': 'en',
        'dimensions': 768,
        'name': 'bert-base-nli-stsb-mean-tokens'
    },
    'en_bert_large_nli_stsb_mean_tokens': {
        'spacy_base_model': 'en',
        'dimensions': 1024,
        'name': 'bert-large-nli-stsb-mean-tokens'
    },
    'en_roberta_base_nli_stsb_mean_tokens': {
        'spacy_base_model': 'en',
        'dimensions': 768,
        'name': 'roberta-base-nli-stsb-mean-tokens'
    },
    'en_roberta_large_nli_stsb_mean_tokens': {
        'spacy_base_model': 'en',
        'dimensions': 1024,
        'name': 'roberta-large-nli-stsb-mean-tokens'
    },
    'en_distilbert_base_nli_stsb_mean_tokens': {
        'spacy_base_model': 'en',
        'dimensions': 768,
        'name': 'distilbert-base-nli-stsb-mean-tokens'
    },
    'xx_distiluse_base_multilingual_cased': {
        'spacy_base_model': 'xx',
        'dimensions': 512,
        'name': 'distiluse-base-multilingual-cased'
    },
    'xx_xlm_r_base_en_ko_nli_ststb': {
        'spacy_base_model': 'xx',
        'dimensions': 768,
        'name': 'xlm-r-base-en-ko-nli-ststb'
    },
    'xx_xlm_r_large_en_ko_nli_ststb': {
        'spacy_base_model': 'xx',
        'dimensions': 1024,
        'name': 'xlm-r-large-en-ko-nli-ststb'
    },
}

def create_lang(model_name):
    '''Creates a Language object from the `model_name`'''
    from . import language
    if model_name not in configs:
        raise ValueError(f'Model "{model_name}" not available')
    selected_config = configs[model_name]
    nlp = language.SentenceBert.create_nlp(selected_config)
    nlp.vocab.reset_vectors(width=selected_config['dimensions'])
    with open(Path(__file__).parent.absolute() / 'meta' / f'{model_name}.json') as f:
        nlp.meta = json.load(f)
    return nlp

def name_spacy_to_sentencebert(spacy_name):
    # remove initial prefix
    # result = spacy_name[3:]
    # from underscore to dash
    result = spacy_name.replace('_', '-')
    return result



# coding: utf8
from __future__ import unicode_literals

import os
from pathlib import Path
import spacy
from spacy.util import load_model_from_init_py, get_model_meta
from spacy.language import Language
from spacy.tokens import Span
from spacy.matcher import Matcher
import warnings
from . import util
from .util import create_lang
from sentence_transformers import SentenceTransformer

from .util import create_lang as load_model

__version__ = util.pkg_meta["version"]


from .language import SentenceBert
SentenceBert.install_extensions()

# warning suppress for empty vocabulary
warnings.filterwarnings('ignore', message=r"\[W007\]", category=UserWarning)

Language.factories['sentencebert_overwrite_vectors'] = lambda nlp, **cfg: OverwriteVectors(nlp, **cfg)
Language.factories['sentencebert_add_model_to_doc'] = lambda nlp, **cfg: AddModelToDoc(nlp, **cfg)

def load(**overrides):
    return load_model_from_init_py(__file__, **overrides)


def create_from(nlp, model_name):
    '''From an existing `nlp` object, adds the vectors from the specific `model_name` by adding pipeline stages'''
    if model_name not in util.configs:
        raise ValueError(f'Model "{model_name}" not available')
    config = util.configs[model_name]
    return SentenceBert.create_nlp(config, nlp)


class OverwriteVectors(object):
    '''Factory of the `sentencebert_overwrite_vectors` pipeline stage. It tells spacy how to create the stage '''
    name = 'sentencebert_overwrite_vectors'

    def __init__(self, nlp):
        # TODO could add a cache, configurable, to store already computed vectors
        # enable_cache = cfg.get('enable_cache', True)
        pass

    def __call__(self, doc):
        SentenceBert.overwrite_vectors(doc)
        return doc

class AddModelToDoc(object):
    '''Factory of the `sentencebert_add_model_to_doc` pipeline stage. It tells spacy how to create the stage '''
    name = 'sentencebert_add_model_to_doc'

    def __init__(self, nlp):
        spacy_name = nlp.meta['name']
        model_name = util.name_spacy_to_sentencebert(spacy_name)
        self.model = SentenceTransformer(model_name)


    def __call__(self, doc):
        doc._.sentence_bert_model = self.model
        return doc



import spacy

from spacy.tokens import Doc, Span, Token
from sentence_transformers import SentenceTransformer

class SentenceBert(object):

    @staticmethod
    def install_extensions():
        '''Creates the extensions on docs, spans, tokens'''
        def vectorise(sent):
            return sent.doc._.sentence_bert_model.encode([sent.text])[0]
        
        # create an extension where the model will be used
        Doc.set_extension('sentence_bert_model', default=None, force=True)

        # set the extension both on doc and span level. This will contain the computed vector
        Token.set_extension('sentence_bert', getter=vectorise, force=True)
        Span.set_extension('sentence_bert', getter=vectorise, force=True)
        Doc.set_extension('sentence_bert', getter=vectorise, force=True)

    
    @staticmethod
    def overwrite_vectors(doc):
        '''Pipeline component that overwrites the vectors'''
        doc.user_hooks["vector"] = lambda a: a._.sentence_bert
        doc.user_span_hooks["vector"] = lambda a: a._.sentence_bert
        doc.user_token_hooks["vector"] = lambda a: a._.sentence_bert
        return doc


    @staticmethod
    def create_nlp(config, nlp=None):
        model = SentenceTransformer(config['name'])

        def add_model_to_doc(doc):
            doc._.sentence_bert_model = model
            return doc
        
        if not nlp:
            nlp = spacy.blank(config['spacy_base_model'])
        nlp.add_pipe(add_model_to_doc, name='sentencebert_add_model_to_doc', first=True)
        nlp.add_pipe(SentenceBert.overwrite_vectors, name='sentencebert_overwrite_vectors', after='sentencebert_add_model_to_doc')

        return nlp
